import collections.abc
import functools
import re
import weakref


UNDEFINED = object()


def _split_path(path, _regex=re.compile(r'(?<!\\)/')):
    """
    Example:

        >>> _split_path('a/b/c')
        ['a', 'b', 'c']
        >>> _split_path('a/b/c/')
        ['a', 'b', 'c', '']
        >>> _split_path('/a/b/c')
        ['', 'a', 'b', 'c']
        >>> _split_path('//')
        ['', '', '']
        >>> _split_path('a\\/b/c')
        ['a/b', 'c']

    """
    return list(map(functools.partial(re.sub, r'\\/', '/'), _regex.split(path)))


class Config(collections.abc.Mapping):
    """Read only mapping-like for holding configuration.

    Note: At the opposite of ``ConfigParser.RawConfigParser`` in stdlib, methods
    like ``get_<type>`` in this class do **not** coerce, instead the use case
    is type assertion, i.e. When a user call ``Config.get_int`` on a key where
    the value is a string e.g. '2', the call will fail with a ``TypeError``.

    """

    def __init__(self, data):
        if not isinstance(data, collections.Mapping):
            raise TypeError('Need a mapping-like object, instead got %r' % type(data))
        self._data = data
        self.__inner_configs = weakref.WeakValueDictionary()

    def __str__(self):
        return '%s(%r)' % (self.__class__.__name__, self._data)

    def _invalidate(self, new_data):
        self._data = new_data
        for path, inner_config in self.__inner_configs.items():
            inner_data = self.get(path, default={})
            inner_config._invalidate(inner_data)

    def __iter__(self):
        return iter(self._data)

    def __len__(self):
        return len(self._data)

    def __getitem__(self, path):
        data = self._data
        keys = _split_path(path)
        for i, key in enumerate(keys):
            try:
                data = data[key]
            except (KeyError, TypeError):
                raise KeyError('No key %r under path %r' % (key, '.'.join(keys[:i])))
        return data

    def get(self, path, default=None, type_=None):
        """Same as ``dict.get(k[,d])`` with an extra argument ``type_`` that assert
        type matching if given.

        ``path`` argument can be in the form 'key1/key2/key3' as a shortcut for doing
        ``config['key1']['key2']['key3']``. In case a key have a / in it, we can escape it
        like so 'key1\\/key2/key3' this will be translated to ``config['key1/key2']['key3']``

        :param path: path expression e.g. 'key1/key2/key3'
        :param default: default value to return when key is missing, in case it's
            supplied ``default`` with ``type_``, ``default`` must be of of type ``type_``,
            default: None.
        :param type_: Python type to validate returned value, default: None.

        :return: the value of the requested path as a string.

        :raises TypeError: if path value is not of type ``type_``.
        :raises KeyError: if path doesn't exist and ``default`` is set to ``distconfig.config.UNDEFINED``.
        """
        if default is UNDEFINED:
            result = self[path]
        else:
            result = super(Config, self).get(path, default=default)
        if type_ and not isinstance(result, type_):
            raise TypeError('%r not of type %r' % (result, type_))
        return result

    def get_config(self, path, default=UNDEFINED):
        """Get path value as :class:`Config`.

        :param path: path expression e.g. 'key1.key2.key3'
        :param default: default value to return when key is missing,
            if supplied ``default`` must be of type dict, default: UNDEFINED.

        :return: the value of the requested path as a :class:`Config` instance.

        :raises TypeError: if path value is not of type ``type_``.
        :raises KeyError: if path doesn't exist and no default was given.
        """
        try:
            instance = self.__inner_configs[path]
        except KeyError:
            inner = self.get(path, default=default, type_=dict)
            instance = self.__inner_configs[path] = Config(inner)
        return instance

    def get_int(self, path, default=UNDEFINED):
        """Get path value as integer.

        :param path: path expression e.g. 'key1.key2.key3'
        :param default: default value to return when key is missing,
            if supplied ``default`` must be of type integer, default: UNDEFINED.

        :return: the value of the requested path as a Integer.

        :raises TypeError: if path value is not of type ``type_``.
        :raises KeyError: if path doesn't exist and no default was given.
        """
        return self.get(path, default=default, type_=int)

    def get_float(self, path, default=UNDEFINED):
        """Get path value as float.

        :param path: path expression e.g. 'key1.key2.key3'
        :param default: default value to return when key is missing,
            if supplied ``default`` must be of type float, default: UNDEFINED.

        :return: the value of the requested path as a Float.

        :raises TypeError: if path value is not of type ``type_``.
        :raises KeyError: if path doesn't exist and no default was given.
        """
        return self.get(path, default=default, type_=float)

    def get_boolean(self, path, default=UNDEFINED):
        """Get path value as boolean.

        :param path: path expression e.g. 'key1.key2.key3'
        :param default: default value to return when key is missing,
            if supplied ``default`` must be of type boolean, default: UNDEFINED.

        :return: the value of the requested path as a Boolean.

        :raises TypeError: if path value is not of type ``type_``.
        :raises KeyError: if path doesn't exist and no default was given.
        """
        return self.get(path, default=default, type_=bool)

    def get_string(self, path, default=UNDEFINED):
        """Get path value as string.

        :param path: path expression e.g. 'key1.key2.key3'
        :param default: default value to return when key is missing,
            if supplied ``default`` must be of type ustring, default: UNDEFINED.

        :return: the value of the requested path as a String.

        :raises TypeError: if path value is not of type ``type_``.
        :raises KeyError: if path doesn't exist and no default was given.
        """
        return self.get(path, default=default, type_=str)

    def get_bytes(self, path, default=UNDEFINED):
        """Get path value as bytes.

        :param path: path expression e.g. 'key1.key2.key3'
        :param default: default value to return when key is missing,
            if supplied ``default`` must be of type bytes, default: UNDEFINED.

        :return: the value of the requested path as Bytes.

        :raises TypeError: if path value is not of type ``type_``.
        :raises KeyError: if path doesn't exist and no default was given.
        """
        return self.get(path, default=default, type_=bytes)





import unittest

from distconfig.config import Config


class ConfigTestCase(unittest.TestCase):

    def setUp(self):
        self.test_data = {
            'inner': {
                'integer': 1,
                'float': 1.4,
                'boolean': True,
                'string': 'Straße',
                'bytes': b'bytes'
            },
            'outer': 'foobar',
            'esca/ped': 'escaped'
        }
        self.config = Config(self.test_data)

    def test_config_wrong_type(self):
        with self.assertRaises(TypeError):
            Config('foobar')

    def test_itemgetter(self):
        self.assertEqual(self.config['outer'], 'foobar')

    def test_get(self):
        self.assertEqual(self.config.get('outer'), 'foobar')

    def test_get_default(self):
        value = self.config.get('inner/noexistant', 'foobar')
        self.assertEqual(value, 'foobar')

    def test_get_escaped_key(self):
        value = self.config.get('esca\\/ped')
        self.assertEqual(value, 'escaped')

    def test_no_key_exception(self):
        with self.assertRaisesRegexp(KeyError, "No key 'unexistant' under path 'inner.boolean'"):
            self.config['inner/boolean/unexistant']

    def test_config_len(self):
        self.assertEqual(len(self.config), 3)

    def test_config_iter(self):
        self.assertEqual(sorted(self.config), ['esca/ped', 'inner', 'outer'])

    def test_contains_path(self):
        self.assertTrue('inner/string' in self.config)
        self.assertTrue('outer' in self.config)

        self.assertTrue('inner/not_there' not in self.config)

    def test_config_str(self):
        self.assertEqual(str(self.config), 'Config(%r)' % self.test_data)

    def test_get_config(self):
        inner_config = self.config.get_config('inner')

        self.assertEqual(inner_config, Config(self.test_data['inner']))

    def test_get_config_with_default(self):
        self.assertEqual(self.config.get_config('inner/not_there', default={}), Config({}))

    def test_get_config_wrong_type(self):
        with self.assertRaises(TypeError):
            self.config.get_config('outer')

        with self.assertRaises(TypeError):
            self.config.get_config('inner/not_there', default=1)

    def test_get_int(self):
        self.assertEqual(self.config.get_int('inner/integer'), 1)

    def test_get_int_with_default(self):
        self.assertEqual(self.config.get_int('inner/not_there', default=1), 1)

    def test_get_int_wrong_type(self):
        with self.assertRaises(TypeError):
            self.config.get_int('outer')

        with self.assertRaises(TypeError):
            self.config.get_int('inner/not_there', default='wrong_type')

    def test_get_float(self):
        self.assertEqual(self.config.get_float('inner/float'), 1.4)

    def test_get_float_with_default(self):
        self.assertEqual(self.config.get_float('inner/not_there', default=2.1), 2.1)

    def test_get_float_wrong_type(self):
        with self.assertRaises(TypeError):
            self.config.get_float('outer')

        with self.assertRaises(TypeError):
            self.config.get_float('inner/not_there', default=False)

    def test_get_bool(self):
        self.assertEqual(self.config.get_boolean('inner/boolean'), True)

    def test_get_bool_with_default(self):
        self.assertEqual(self.config.get_boolean('inner/not_there', default=False), False)

    def test_get_boolean_wrong_type(self):
        with self.assertRaises(TypeError):
            self.config.get_boolean('outer')

        with self.assertRaises(TypeError):
            self.config.get_boolean('inner/not_there', default='wrong type')

    def test_get_string(self):
        self.assertEqual(self.config.get_string('inner/string'), 'Straße')

    def test_get_string_with_default(self):
        self.assertEqual(self.config.get_string('inner/not_there', default=''), '')

    def test_get_string_wrong_type(self):
        with self.assertRaises(TypeError):
            self.config.get_string('inner/integer')

        with self.assertRaises(TypeError):
            self.config.get_string('inner/not_there', default=1)

    def test_get_bytes(self):
        self.assertEqual(self.config.get_bytes('inner/bytes'), b'bytes')

    def test_get_bytes_with_default(self):
        self.assertEqual(self.config.get_bytes('inner/not_there', default=b''), b'')

    def test_get_bytes_wrong_type(self):
        with self.assertRaises(TypeError):
            self.config.get_bytes('inner/string')

        with self.assertRaises(TypeError):
            self.config.get_bytes('inner/not_there', default=1)


class InnerConfigTestCase(unittest.TestCase):

    def setUp(self):
        self.test_data = {
            'inner': {
                'foo': 'bar',
                'subinner': {
                    'baz': 'taz'
                }
            },
            'another_inner': {
                'foobar': 'wat?'
            }
        }
        self.config = Config(self.test_data)

    def test_inner_config_update(self):
        inner_config = self.config.get_config('inner')
        self.config._invalidate({
            'inner': {
                'integer': 5
            }
        })

        self.assertEqual(inner_config.get('integer'), 5)

    def test_two_inners_update(self):
        inner = self.config.get_config('inner')
        another_inner = self.config.get_config('another_inner')

        self.config._invalidate({
            'inner': {}
        })

        self.assertEqual(inner._data, {})
        self.assertEqual(another_inner._data, {})

    def test_nested_inner_config_update(self):
        inner = self.config.get_config('inner')
        subinner = self.config.get_config('inner/subinner')

        self.config._invalidate({
            'inner': {
                'new': 'yay',
                'subinner': {
                    'subnew': 'subyay'
                }
            }
        })

        self.assertEqual(inner.get('new'), 'yay')
        self.assertEqual(inner.get('subinner/subnew'), 'subyay')
        self.assertEqual(subinner.get('subnew'), 'subyay')

    def test_identity_inner_config(self):
        inner1 = self.config.get_config('inner')
        inner2 = self.config.get_config('inner')

        self.assertIs(inner1, inner2)
